import unittest

from canonical_args import structure



class TestCase_CheckList(unittest.TestCase):

	def test_non_nested(self):
		names = ["arg1", "arg2", "arg3"]
		types = [int, float, str]
		values = [None, None, None]
		args = [1, 13.2, "thing"]

		structure.check_list(names, types, values, args)

	def test_non_nested_fail(self):
		names = ["arg1", "arg2", "arg3"]
		types = [int, float, str]
		values = [None, None, None]
		args = [1.1, 13.2, "thing"]

		try:
			structure.check_list(names, types, values, args)
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")

	def test_nested_list(self):
		names = ["arg1", "arg2", "arg3"]
		types = [int, float, [str, bool]]
		values = [None, None, [["thing1", "thing2"], None]]
		args = [1, 13.2, ["thing1", False]]

		structure.check_list(names, types, values, args)

	def test_nested_list_fail(self):
		names = ["arg1", "arg2", "arg3"]
		types = [int, float, [str, bool]]
		values = [None, None, [["thing1", "thing2"], None]]
		args = [1, 13.2, ["thing1"]]  # nested list is wrong size

		try:
			structure.check_list(names, types, values, args)
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")

	def test_nested_dict(self):
		names = ["arg1", "arg2", "arg3"]
		types = [int, float, {"a": {"type": int, "values": None},
							  "b": {"type": float, "values": None}}]
		values = [None, None, None]
		args = [1, 13.2, {"a": 15, "b": 22.13}]

		structure.check_list(names, types, values, args)

	def test_nested_dict_fail(self):
		names = ["arg1", "arg2", "arg3"]
		types = [int, float, {"a": {"type": int, "values": None},
							  "b": {"type": float, "values": None}}]
		values = [None, None, None]
		args = [1, 13.2, {"a": 15}]  # missing key "b"

		try:
			structure.check_list(names, types, values, args)
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")


class TestCase_CheckDict(unittest.TestCase):

	def test_non_nested(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": float,
				"values": None
			}
		}
		kwargs = {"kwarg1": 1, "kwarg2": 33.1}

		structure.check_dict(spec, kwargs)

	def test_non_nested_fail(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": float,
				"values": None
			}
		}
		kwargs = {"kwarg1": 1, "kwarg2": 3}  # kwarg2 has bad type

		try:
			structure.check_dict(spec, kwargs)
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")

	def test_non_nested_non_required(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": float,
				"values": None,
				"required": False
			}
		}
		kwargs = {"kwarg1": 1}  # non required kwarg missing

		structure.check_dict(spec, kwargs)

	def test_non_nested_required(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": float,
				"values": None,
				"required": True
			}
		}
		kwargs = {"kwarg1": 1, "kwarg2": 3.3}

		structure.check_dict(spec, kwargs)

	def test_non_nested_required_fail(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": float,
				"values": None,
				"required": True
			}
		}
		kwargs = {"kwarg1": 1}  # missing required kwarg2

		try:
			structure.check_dict(spec, kwargs)
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")


	def test_nested_list(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": [int, int, float],
				"values": [None, None, None]
			}
		}
		kwargs = {"kwarg1": 1, "kwarg2": [1, 2, 3.5]}

		structure.check_dict(spec, kwargs)

	def test_nested_list_fail(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": [int, int, float],
				"values": [None, None, None]
			}
		}
		kwargs = {"kwarg1": 1, "kwarg2": [1, 2]}  # bad length

		try:
			structure.check_dict(spec, kwargs)
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")

	def test_nested_dict(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": {
					"subkey1": {
						"type": str,
						"values": None
					},
					"subkey2": {
						"type": int,
						"values": None
					}
				},
				"values": None
			}
		}
		kwargs = {"kwarg1": 1, "kwarg2": {"subkey1": "a", "subkey2": 1}}

		structure.check_dict(spec, kwargs)

	def test_nested_dict_fail(self):
		spec = {
			"kwarg1": {
				"type": int,
				"values": None
			},
			"kwarg2": {
				"type": {
					"subkey1": {
						"type": str,
						"values": None
					},
					"subkey2": {
						"type": int,
						"values": None
					}
				},
				"values": None
			}
		}
		kwargs = {"kwarg1": 1, "kwarg2": {"subkey1": "a"}}  # missing key

		try:
			structure.check_dict(spec, kwargs)
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")

class TestCase_CheckArgs(unittest.TestCase):

	def test_check_args(self):
		spec = {
			"args": [
				{
					"name": "arg1",
					"type": int,
					"values": None
				}
			],
			"kwargs": {
				"kwarg1": {
					"type": float,
					"values": None
				}
			}
		}

		structure.check_args(spec, [1], {"kwarg1": 33.1})

	def test_check_args_fail(self):
		spec = {
			"args": [
				{
					"name": "arg1",
					"type": int,
					"values": None
				}
			],
			"kwargs": {
				"kwarg1": {
					"type": float,
					"values": None
				}
			}
		}

		try:
			structure.check_args(spec, [1], {"kwarg1": "string"})  # bad type
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")



"""
Provides utilities for checking the structure of a list
or dict against provided structure mandates and details.
"""
from __future__ import absolute_import

import json

from . import check



def check_list(names, types, values, arg):
    """
    recursively check a list of argument values (``arg``).
    usually used to check positional arguments passed to a
    function call.
    """
    # length
    try:
        assert len(types) == len(arg)
    except AssertionError as e:
        raise AssertionError("expected {} positional arguments, only"\
                             " got {}".format(len(types), len(arg)))

    for subname, subtype, subvalues, subarg in zip(names, types, values, arg):

        subtype = check.eval_subtype(subtype)
        
        # recurse if list or tuple, but not if choice of one
        if ((isinstance(subtype, list) or
             isinstance(subtype, tuple)) and
            not isinstance(subtype, check.ChoiceOfOne)):
            subname = ["{}#{}".format(subname, i) \
                       for i in range(0, len(subtype))]
            check_list(subname, subtype, subvalues, subarg)
            continue

        elif isinstance(subtype, dict):
            check_dict(subtype, subarg)
            continue

        subarg = check.check_subtype(subname, subtype, subarg)
        check.check_value(subname, subtype, subvalues, subarg)


def check_dict(structure_dict, kwargs):
    """
    usually used for checking kwargs
    """
    required = [key for key, spec in structure_dict.items()\
                    if "required" not in spec or spec["required"]]

    # keys may not be missing
    missing = set(required) - set(kwargs.keys())
    assert len(missing) == 0

    for subname, subarg in kwargs.items():
        # make sure the keyword argument is registered
        assert subname in structure_dict
        struct = structure_dict[subname]

        subtype = check.eval_subtype(struct["type"])

        if ((isinstance(subtype, list) or
             isinstance(subtype, tuple)) and
            not isinstance(subtype, check.ChoiceOfOne)):
            subname = ["{}#{}".format(subname, i) for i in range(0, len(subtype))]
            check_list(subname, subtype, struct["values"], subarg)
            continue

        elif isinstance(subtype, dict):
            check_dict(subtype, subarg)
            continue

        subarg = check.check_subtype(subname, subtype, subarg)
        check.check_value(subname, subtype, struct["values"], subarg)


def check_args(struct, args, kwargs):
    """
    check the arguments provided against the structure.

    :param dict struct: the dict detailing the argument structure
    :param list args: the args to check against struct
    :param dict kwargs: the kwargs to check against the struct
    :raises AssertionError: when args or kwargs do not match struct
    """
    names = [sub["name"] for sub in struct["args"]]
    types = [sub["type"] for sub in struct["args"]]
    values = [sub["values"] for sub in struct["args"]]
    # args = args

    check_list(names, types, values, args)
    if kwargs:  # kwargs are optional!
        check_dict(struct["kwargs"], kwargs)




"""
Provides decorators and utilities for interfacing
``canonical_args`` with python methods.
"""
from __future__ import absolute_import

import functools
import json
import types

from . import structure



FUNC_SPEC_VAR = "__canonical_arg_spec"
GET_SPEC_FUNC = "get_spec"

def arg_spec(spec, register=True):
    """
    Decorates a method, and checks args against spec.

    :params dict spec: the arg spec dict
    :params bool register: default ``True``, registers the spec
        directly to the function variable scope. once registered,
        the spec can be retrieved with ``func.get_spec()``.
    """
    def inner(func):
        # register the spec to the function,
        if register and not hasattr(func, FUNC_SPEC_VAR):
            register_spec(func, spec)

        @functools.wraps(func)
        def _inner(*args, **kwargs):
            if register:
                spec = getattr(func, FUNC_SPEC_VAR)
            structure.check_args(spec, args, kwargs)
            return func(*args, **kwargs)
        return _inner
    return inner

def register_spec(func, spec):
    """
    add the argspec var to the function scope
    """
    setattr(func, FUNC_SPEC_VAR, spec)
    setattr(func, GET_SPEC_FUNC, types.MethodType(get_spec, func, type(func)))

def get_spec(func):
    """
    instance method inserted dynamically into a method scope,
    allowing retrieval of the arg spec associated with the message: ::

        >>> @arg_spec({...}, register=True)
        ... def f(x):
        ...     return x
        ...
        >>> function.get_spec()
        {...}

    Can also be used directly, as in: ::

        >>> function.get_spec(some.function.name)
        {...}

    """
    if not hasattr(func, FUNC_SPEC_VAR):
        raise AttributeError("{} does not have a registered spec."\
                             " Consider using the ``register=True``"\
                             " flag in the ``arg_spec`` decorator.".format(
            func))
    return getattr(func, FUNC_SPEC_VAR)



import unittest

from canonical_args import function


spec = {
	"args": [
		{"name": "arg1", "type": int, "values": None},
		{"name": "arg2", "type": float, "values": None}
	],
	"kwargs": {
		"kwarg1": {"type": int, "values": None}
	}
}
@function.arg_spec(spec, register=True)
def func1(*args, **kwargs):
	return args, kwargs

class TestCase_TestRegisterToFunc(unittest.TestCase):

	def test_get_func_scope(self):
		self.assertEqual(getattr(func1, function.FUNC_SPEC_VAR), spec)

	def test_get_func_spec_with_method(self):
		self.assertEqual(func1.get_spec(), spec)

class TestCase_TestArgSpecDecorator(unittest.TestCase):

	def test_fire_decorated_method(self):
		func1(1, 3.1, kwarg1=1)

	def test_fire_decorated_method_fail(self):
		try:
			func1(1, 3.1, kwarg1="string")  # bad kwarg1 type
		except AssertionError:
			pass
		else:
			self.fail("should have thrown AssertionError")



"""
Make top-level imports available.
"""
from __future__ import absolute_import

from .structure import check_args
from .function import arg_spec, register_spec



__all__ = ["check_args",
		   "arg_spec",
		   "register_spec"]



"""

"""
from __future__ import absolute_import



class ChoiceOfOne(list):
    pass

def dynamic_import(class_string):
    """
    perform a dynamic import of a class or function (callable).
    """
    package_name = class_string.split(".")[0]

    mod = __import__(package_name)
    for comp in class_string.split(".")[1:]:
        mod = getattr(mod, comp)
    return mod

def eval_subtype(subtype):
    """
    get the subtype from a subtype string.

    :param subtype: the string version of the type
    :type subtype: ``str`` or ``unicode``

    Acceptable values:

    =========      =====================================================
    Format         Use
    =========      =====================================================
    "int"          native types: str, float, int, bool, dict, list, etc.
    "one([])"      choice of one, can nest 'cls' calls here
    "cls('')"      a class import string
    =========      =====================================================
    """
    def one(subtype):
        assert isinstance(subtype, list)
        return ChoiceOfOne(eval_subtype(subtype))

    def cls(subtype):
        assert isinstance(subtype, str) or isinstance(subtype, unicode)
        # subtype = re.findall("\((.*)\)", subtype)[0]
        return dynamic_import(subtype)

    if isinstance(subtype, str) or isinstance(subtype, unicode):

        try:
            subtype = eval(subtype)
        except NameError:
            # assume this is a class for backwards compat
            subtype = dynamic_import(subtype)
        except SyntaxError:
            raise SyntaxError("unable to evaluate type string '{}'".format(
                subtype))

    elif isinstance(subtype, list):
        for i in range(0, len(subtype)):
            subtype[i] = eval_subtype(subtype[i])

    return subtype

def check_subtype(subname, subtype, subarg, should_cast=False):
    """

    """
    # choice of one
    if isinstance(subtype, ChoiceOfOne):
        try:
            assert type(subarg) in subtype
        except AssertionError:
            raise AssertionError("expected argument '{}' of types"\
                                 " {}. got {}".format(
                subname, subtype, type(subarg)))

    # straight up type check
    else:
        try:
            assert isinstance(subarg, subtype) or subarg is None
        except AssertionError as e:
            if should_cast:
                try:
                    subarg = subtype(subarg)
                except ValueError as e:
                    raise ValueError("unable to cast argument '{}' of"\
                                     "type {} to type {}".format(
                        subname, type(subarg), subtype))
            else:
                raise AssertionError("expected argument '{}' of type {}."\
                                     " got {}".format(
                    subname, subtype, type(subarg)))

    return subarg

def check_value(subname, subtypes, subvalues, subarg):
    """

    """
    if isinstance(subtypes, ChoiceOfOne):
        subtype = str(type(subarg)).replace("<type '", "").replace("'>", "")
        subtype = subtype.replace("<class '", "").replace("'>", "")
        subvalues = subvalues[subtype]

    # whitelisting
    check_value_whitelist(subname, subvalues, subarg)

    # number ranging (int and float)
    check_value_range(subname, subvalues, subarg)

    # number gt, lt, gte, lte
    check_value_comparison(subname, subvalues, subarg)

def check_value_whitelist(subname, subvalues, subarg):
    """

    """
    if isinstance(subvalues, list):
        try:
            assert subarg in subvalues
        except AssertionError as e:
            raise AssertionError("`{}` is not a permitted value for"\
                                 " arg '{}'. expected value from"\
                                 " {}".format(subarg, subname, subvalues))

def check_value_range(subname, subvalues, subarg):
    """

    """
    # number ranging (int and float)
    if ((isinstance(subvalues, str) or
         isinstance(subvalues, unicode)) and
        subvalues.startswith("range")):
        
        ran = eval(subvalues.replace("range", ""))
        try:
            assert ran[0] <= subarg and subarg <= ran[-1]
        except AssertionError as e:
            raise AssertionError("`{}` is not in range({}, {}) for"\
                                 " arg '{}'".format(subarg,
                                                    ran[0],
                                                    ran[-1],
                                                    subname))

def check_value_comparison(subname, subvalues, subarg):
    """

    """
    if ((isinstance(subvalues, str) or
         isinstance(subvalues, unicode)) and
        (subvalues.startswith(">") or subvalues.startswith("<"))):

        ts = "{}{}".format(subarg, subvalues)
        try:
            assert eval(ts)
        except AssertionError as e:
            raise AssertionError("`{}` is not {} for arg {}".format(
                subarg, subvalues, subname))




from setuptools import setup, find_packages


description = "A module for dynamically registering, checking, and"\
			  " working with nested argument specs."
long_description = open("README.rst").read()

setup(name="canonical_args",
	  version="0.3",
	  description=description,
	  long_description=long_description,
	  author="Shonte Amato-Grill",
	  author_email="shonte.amatogrill@gmail.com",
	  maintainer="Shonte Amato-Grill",
	  maintainer_email="shonte.amatogrill@gmail.com",
	  url="https://github.com/shonteag/canonical_args",
	  packages=find_packages(exclude=["test", "tests"]),
	  classifiers=[
	      "Programming Language :: Python :: 2.7",
	      "Topic :: Software Development :: Libraries :: Python Modules",
	      "Intended Audience :: Developers"
	  ]
)
